
/**************************************************************************
 *
 *  FILE            shhlog.c
 *
 *  DESCRIPTION     Contains routines for logging to file or pipe.
 *
 *  WRITTEN BY      Sverre H. Huseby
 *                  Hoffsvn. 50 leil. 105
 *                  N-0377 Oslo
 *                  Norway
 *
 *                  sverrehu@ifi.uio.no
 *
 **************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <time.h>
#include <fcntl.h>
#include <unistd.h>

#ifndef SEEK_SET
  #define SEEK_SET 0
  #define SEEK_CUR 1
  #define SEEK_END 2
#endif

#include "shhlog.h"

/* Define WANT_LOCKING to include file-locking */
/* #define WANT_LOCKING /**/

extern int errno;

/**************************************************************************
 *                                                                        *
 *                       P R I V A T E    D A T A                         *
 *                                                                        *
 **************************************************************************/

static FILE *logStream = NULL;
static int logWhat = 0, isPipe = 0;





/**************************************************************************
 *                                                                        *
 *                   P R I V A T E    F U N C T I O N S                   *
 *                                                                        *
 **************************************************************************/

#ifdef WANT_LOCKING
/*-------------------------------------------------------------------------
 *
 *  NAME          lockStream
 *
 *  FUNCTION      Lock the stream.
 *
 *  RETURNS       -1 on error, 0 if ok.
 *
 */
static int lockStream(void)
{
    struct flock fl;

    fl.l_type = F_WRLCK;
    fl.l_whence = SEEK_SET;
    fl.l_start = 0L;
    fl.l_len = 0L;
    return fcntl(fileno(logStream), F_SETLKW, fl);
}



/*-------------------------------------------------------------------------
 *
 *  NAME          unlockStream
 *
 *  FUNCTION      Unlock the stream.
 *
 *  RETURNS       -1 on error, 0 if ok.
 *
 */
static int unlockStream(void)
{
    struct flock fl;

    fl.l_type = F_UNLCK;
    fl.l_whence = SEEK_SET;
    fl.l_start = 0L;
    fl.l_len = 0L;
    return fcntl(fileno(logStream), F_SETLK, fl);
}
#endif



/**************************************************************************
 *                                                                        *
 *                    P U B L I C    F U N C T I O N S                    *
 *                                                                        *
 **************************************************************************/

/*-------------------------------------------------------------------------
 *
 *  NAME          logOpen
 *
 *  FUNCTION      Open file or pipe for logging.
 *
 *  SYNOPSIS      #include "shhlog.h"
 *                int logOpen(const char *name, int what);
 *
 *  INPUT         name    Name of file to log to. If the string starts with
 *                        a bar (|), the rest of the string is taken as
 *                        a command, and a pipe is opened to it.
 *                what    What to prepend to each logged line. Any
 *                        combination of the logXXX-constants.
 *
 *  RETURNS       0 if ok, -1 on error. Use errno to check.
 *
 */
int logOpen(const char *name, int what)
{
    logWhat = what;
    if (*name == '|') {
	/* Open a pipe... */
	logStream = popen(name + 1, "w");
	isPipe = 1;
    } else {
	/* Open a file */
	logStream = fopen(name, "at");
	isPipe = 0;
    }
    return logStream ? 0 : -1;
}



/*-------------------------------------------------------------------------
 *
 *  NAME          logSetStream
 *
 *  FUNCTION      Give the stream to use for logging.
 *
 *  SYNOPSIS      #include "shhlog.h"
 *                int logSetStream(const FILE *stream, int what);
 *
 *  INPUT         name    The stream to use.
 *                what    What to prepend to each logged line. Any
 *                        combination of the logXXX-constants.
 *
 *  RETURNS       0 if ok, -1 on error. Use errno to check.
 *
 */
int logSetStream(FILE *stream, int what)
{
    logWhat = what;
    logStream = stream;
    isPipe = 0;
    return 0;
}



/*-------------------------------------------------------------------------
 *
 *  NAME          logLine
 *
 *  FUNCTION      Put given line in the log.
 *
 *  SYNOPSIS      #include "shhlog.h"
 *                void logLine(const char *format, ...);
 *
 *  INPUT         format, ...
 *                        Arguments used as with printf().
 *
 *  RETURNS       0 if ok, -1 on error. Use errno to check.
 *
 *  DESCRIPTION   Nothing is logged if the logfile is not open. An unopened
 *                logfile is not taken to be an error, so 0 is returned.
 *
 */
int logLine(const char *format, ...)
{
    va_list ap;
    char pre[40], tmp[20], tmp2[20], *p;
    struct tm *tm;
    time_t t;
    static char hostname[20];

    if (!logStream)
        return 0;

    *pre = '\0';
    if (logWhat & (logDATE | logTIME)) {
	time(&t);
	tm = localtime(&t);
	if (logWhat & logDATE) {
	    strftime(tmp, 20, "%x", tm);
	    sprintf(tmp2, "%s%s", *pre ? " " : "", tmp);
	    strcat(pre, tmp2);
	}
	if (logWhat & logTIME) {
	    strftime(tmp, 20, "%X", tm);
	    sprintf(tmp2, "%s%s", *pre ? " " : "", tmp);
	    strcat(pre, tmp2);
	}
    }
    if (logWhat & logPID) {
	sprintf(tmp, "%s[%d]", *pre ? " " : "", (int) getpid());
	strcat(pre, tmp);
    }
    if (logWhat & logHOST) {
	if (!*hostname) {
	    gethostname(hostname, 20);
	    if ((p = strchr(hostname, '.')) != NULL)
		*p = '\0';
	}
	sprintf(tmp, "%s%s", *pre ? " " : "", hostname);
	strcat(pre, tmp);
    }
    if (logWhat)
	strcat(pre, ": ");

  #ifdef WANT_LOCKING
    lockStream();
  #endif
    fflush(logStream);
    if (!isPipe) 
	fseek(logStream, 0L, SEEK_END);
    fputs(pre, logStream);
    va_start(ap, format);
    vfprintf(logStream, format, ap);
    va_end(ap);
    fflush(logStream);
  #ifdef WANT_LOCKING
    unlockStream();
  #endif
    return ferror(logStream) ? -1 : 0;
}



/*-------------------------------------------------------------------------
 *
 *  NAME          logClose
 *
 *  FUNCTION      Close the connection to the log.
 *
 *  SYNOPSIS      #include "shhlog.h"
 *                int logClose(void);
 *
 *  RETURNS       0 if ok, -1 on error. Use errno to check.
 *
 */
int logClose(void)
{
    FILE *f;
    
    if (!logStream)
	return 0;
    f = logStream;
    logStream = NULL;
    return fclose(f);
}
